/*  Calculate Greenspan/Barnes/Hut accelerations.			*/
/*  Source file:  accel1.cs		Prototypes:  accel1.hs		*/
/*  Tree should have been constructed and filled as in tree.cs, fill.cs	*/
/*  Uses declarations in gbh.hs.					*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cscomm.h>
#include "gbh.hs"


/*  Calculate accelerations according to error parameters,	*/
/*  theta, beta, tree data Xp0, Xp1, Tm, Th.			*/
/*  Call from Particles with all processors active.		*/
/*  Returns number of cycles of tree traversal
/*  Leaves accelerations in Ac,  statistics on calculation in	*/
/*	nodex - nodes examined and rejected			*/
/*	nodea - nodes examined and accepted			*/
/*	cactv - total active vp cycles
/*  Doubles allow for very large numbers.  Ints are sufficient	*/
/*  for counting in individual vp's.				*/
/*  The (usually N) times that a node examines its own leaf	*/
/*  are ignored in both counts.					*/

#define Tm (*pTm)
#define Th (*pTh)

int Accel1(float:Tree Xp0[3], float:Tree Xp1[3], float:Tree *pTm,
           float:Tree *pTh, float theta, float beta,
	   double *nodex, double *nodea)	/* For analysis */
 { float:Particles D[3],	/* D0 + D1		*/
       r0, r1, rd, v, w;	/* Distances, working	*/
   int:Particles  ar,		/* 1: accept, 2: reject	*/
		  nd,		/* Tree node		*/
            prx, pra;		/* Counters (analysis)	*/
   CMC_sendaddr_t:Particles sn;	/* Address of [nd]Tree	*/
   int acrj, und, unx, u, cyc;

   everywhere
   { prx = pra  = ar = 0;  cyc = 0;	/* Cycle counter */
     for ( u = 0; u < 3; u++ ) Ac[u] = 0.0f;	/* Clear Ac      */
     /* Repeat under host control until particles split up */
     und = 1;	/* Node examined by host */
     do		/* May also exit with break	 */
      { cyc++;  r0 = r1 = rd = 0.0f;
	/* printf("Cycle %d, node %d\n", cyc, und);	    /*** Testing ***/
	for ( u = 0; u < 3; u++ )		/*  Calculate D, r's	*/
	 { v = [und]Xp0[u] - X0[u];  r0 += v*v;
	   D[u] = v += w = [und]Xp1[u] - X1[u];  r1 += w*w;  rd += v*v;
	 }
	r0 = sqrt(r0);  r1 = sqrt(r1);  rd = sqrt(rd);
	ar = 2;  w = r0 + r1;
	/* Acceptance test: */
	where ( ([und]Th <= theta*(r0 <? r1)) & (rd >= beta*w) ) ar = 1;
	/* printf("u ar pra prx w r0 r1 D[0]:\n");	/*** Testing ***/
        /* Adiag(&ar, &pra, &prx, &w, &r0, &r1, &D[0]);	/*** Testing ***/
	acrj = (|= ar);			/* Are all the same?	*/
	if ( acrj == 3 ) break;		/* If not, split up	*/
	if ( acrj == 1 )		/* All accept		*/
	 { /* Multiplier, avoiding particle associated with leaf	*/
	   where ( w > 0.0f )
	    { w = [und]Tm/(r0*r1*w);  pra++; }
	    else  w = 0.0f;
	   for ( u = 0; u < 3; u++ )  Ac[u] += w*D[u];
	   und = [und]Tnext;
	 }
	 else	/* All reject, can't be leaf */
	 { und = [und]Child0;  prx++; }
      }  /* May exit from break with und > 0 */
     while ( und > 0);	/* end (until particles split up)	*/
     /* printf("Exit host loop, und = %d\n", und);	    /*** Testing ***/
     if ( und > 0 )	/* Particles have split, more nodes to examine	*/
      { where ( ar == 1 )	/* Accepted und	*/
	 { where ( w > 0.0f )
	    { w = [und]Tm/(r0*r1*w);  pra++; }
	    else  w = 0.0f;
	   for ( u = 0; u < 3; u++ )  Ac[u] += w*D[u];
	   nd = [und]Tnext;
	 }			/* end accepted und	*/
	 else	/* Rejected und */
	 { nd = [und]Child0;  prx++; }
	/*printf("Particles have split.\n");	    /*** Testing ***/

	/* Loop to finish acceleration, particles act independently	*/
	/* Each particle is ready to examine node nd (-1 if finished)	*/
	/* Not all particles are finished.				*/
	do
	 { cyc++;  r0 = r1 = rd = 0.0f;		/* Everywhere	*/
	   /* printf("Cycle %d\n", cyc);	    /*** Testing ***/
	   where ( nd > 0 )		/* Particle is still active	*/
	    { sn = make_send_address(Tree, nd);
	      for ( u = 0; u < 3; u++ )
	       { v = get(sn, &Xp0[u], CMC_collisions) - X0[u];
		 r0 += v*v;
		 w = get(sn, &Xp1[u], CMC_collisions) - X1[u];
		 D[u] = v += w;  r1 += w*w;  rd += v*v;
	       }
	      r0 = sqrt(r0);  r1 = sqrt(r1);  rd = sqrt(rd);  w = r0 + r1;
	      where ( get(sn, pTh, CMC_collisions) <= theta*(r0 <? r1)
		     & (rd >= beta*w) 				      )
	       { where ( w > 0.0f )
	          { w = get(sn, pTm, CMC_collisions)/(r0*r1*w);  pra++; }
	          else  w = 0.0f;
	         for ( u = 0; u < 3; u++ )  Ac[u] += w*D[u];
	         nd = get(sn, pTnext, CMC_collisions);
	       }
	       else	/* Rejected nd */
	       { nd = get(sn, pChild0, CMC_collisions);  prx++; }
	    }	/* end where ( nd > 0 )	*/
	 }
	while ( (|= (nd > 0)) & cyc <= Nodes );	/* Safety */
      }	/* end if ( und > 0 )	*/
     /* Report statistics */
     *nodex = (+= (double:current)prx);
     *nodea = (+= (double:current)pra);
     return (cyc);
   }
 }
